# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from typing import Optional, Tuple
import torch as pt
from .models.model import TorchMD_Net
from .models.torchmd_gn import TorchMD_GN


class TorchMD_GN_optimized(pt.nn.Module):
    """This Module is equivalent  to TorchMD_GN, but replaces some of
    its internal operation by optimized ones from NNPops.
    """

    def __init__(self, model):
        from NNPOps.CFConv import CFConv
        from NNPOps.CFConvNeighbors import CFConvNeighbors

        if model.rbf_type != "gauss":
            raise ValueError('Only rbf_type="gauss" is supproted')
        if model.trainable_rbf:
            raise ValueError("trainalbe_rbf=True is not supported")
        if model.activation != "ssp":
            raise ValueError('Only activation="ssp" is supported')
        if model.neighbor_embedding:
            raise ValueError("neighbor_embedding=True is not supported")
        if model.cutoff_lower != 0.0:
            raise ValueError("Only lower_cutoff=0.0 is supported")
        if model.aggr != "add":
            raise ValueError('Only aggr="add" is supported')

        super().__init__()
        self.model = model

        self.neighbors = CFConvNeighbors(self.model.cutoff_upper)

        offset = self.model.distance_expansion.offset
        width = offset[1] - offset[0]
        self.convs = [
            CFConv(
                gaussianWidth=width,
                activation="ssp",
                weights1=inter.mlp[0].weight.T,
                biases1=inter.mlp[0].bias,
                weights2=inter.mlp[2].weight.T,
                biases2=inter.mlp[2].bias,
            )
            for inter in self.model.interactions
        ]

    def forward(
        self,
        z: pt.Tensor,
        pos: pt.Tensor,
        batch: pt.Tensor,
        box: Optional[pt.Tensor] = None,
        q: Optional[pt.Tensor] = None,
        s: Optional[pt.Tensor] = None,
    ) -> Tuple[pt.Tensor, Optional[pt.Tensor], pt.Tensor, pt.Tensor, pt.Tensor]:

        assert pt.all(batch == 0)
        assert box is None, "Box is not supported"

        x = self.model.embedding(z)

        self.neighbors.build(pos)
        for inter, conv in zip(self.model.interactions, self.convs):
            y = inter.conv.lin1(x)
            y = conv(self.neighbors, pos, y)
            y = inter.conv.lin2(y)
            y = inter.act(y)
            x = x + inter.lin(y)

        return x, None, z, pos, batch

    def __repr__(self):
        return "Optimized: " + repr(self.model)


def optimize(model):
    """
    Returns an optimized version for a given TorchMD_Net model.
    If the model is not supported, a ValueError is raised.
    """
    assert isinstance(model, TorchMD_Net)

    if isinstance(model.representation_model, TorchMD_GN):
        model.representation_model = TorchMD_GN_optimized(model.representation_model)
    else:
        raise ValueError("Unsupported model! Only TorchMD_GN is suppored.")

    return model



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import torch
import warp as wp

MODULES = {}


def get_module(name: str, dtype: list[str]):
    """Get the module for the given name and dtype."""
    full_name = name + "_" + "_".join(get_dtype(d) for d in dtype)
    if full_name not in MODULES:
        print(f"Module {full_name} not found in MODULES dictionary")
        print(f"Available modules: {list(MODULES.keys())}")
        raise ValueError(f"Module {full_name} not found")
    return MODULES[full_name]


def add_module(name: str, dtype: list[str], kernel: wp.Kernel):
    """Add the module for the given name and dtype."""
    full_name = name + "_" + "_".join(get_dtype(d) for d in dtype)
    if full_name not in MODULES:
        MODULES[full_name] = kernel
    return MODULES[full_name]


def get_dtype(dtype: str):
    """Get the dtype string representation for the given dtype (WIP)."""
    if dtype.endswith("16"):
        return "fp16"
    if dtype.endswith("32"):
        return "fp32"
    if dtype.endswith("64"):
        return "fp64"
    raise ValueError(f"Unsupported dtype: {dtype}")


def get_wp_fp_dtype(dtype: str):
    """Get the warp dtype for the given dtype (WIP)."""
    if dtype.endswith("16"):
        return wp.float16
    if dtype.endswith("32"):
        return wp.float32
    if dtype.endswith("64"):
        return wp.float64
    raise ValueError(f"Unsupported dtype: {dtype}")


def list_modules():
    """List all modules in the MODULES dictionary."""
    print("Available modules:")
    for name in MODULES:
        print(f"  - {name}")
    return list(MODULES.keys())


def get_stream(device: torch.device):
    """Get the stream for the given device."""
    if device.type == "cuda":
        return wp.stream_from_torch(torch.cuda.current_stream(device))
    return None



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import torch
import warp as wp
from torch import Tensor

from torchmdnet.extensions.warp_kernels import get_module, get_stream


@torch.library.custom_op(
    "tensornet::message_passing_fwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(
    x: Tensor,
    y: Tensor,
    z: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indices: Tensor,
    row_indptr: Tensor,
    col_data: Tensor,
    col_indices: Tensor,
    col_indptr: Tensor,
) -> list[Tensor]:
    stream = get_stream(x.device)
    device = wp.device_from_torch(x.device)
    output_x = torch.empty_like(x)
    output_y = torch.empty_like(y)
    output_z = torch.empty_like(z)

    x_wp = wp.from_torch(x.detach(), return_ctype=True)
    y_wp = wp.from_torch(y.detach(), return_ctype=True)
    z_wp = wp.from_torch(z.detach(), return_ctype=True)

    output_x_wp = wp.from_torch(output_x.detach(), return_ctype=True)
    output_y_wp = wp.from_torch(output_y.detach(), return_ctype=True)
    output_z_wp = wp.from_torch(output_z.detach(), return_ctype=True)

    edge_attr_wp = wp.from_torch(edge_attr.detach(), return_ctype=True)

    row_data_wp = wp.from_torch(row_data.detach(), return_ctype=True)
    row_indices_wp = wp.from_torch(row_indices.detach(), return_ctype=True)
    row_indptr_wp = wp.from_torch(row_indptr.detach(), return_ctype=True)

    message_passing_fwd = get_module("message_passing_fwd", [str(x.dtype)])
    wp.launch(
        message_passing_fwd,
        dim=(x.shape[0], x.shape[-1]),
        stream=stream,
        device=device,
        inputs=(
            x_wp,
            y_wp,
            z_wp,
            edge_attr_wp,
            row_data_wp,
            row_indices_wp,
            row_indptr_wp,
            output_x_wp,
            output_y_wp,
            output_z_wp,
        ),
    )

    return [output_x, output_y, output_z]


@torch.library.register_fake("tensornet::message_passing_fwd_primitive")
def _(
    x: Tensor,
    y: Tensor,
    z: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indices: Tensor,
    row_indptr: Tensor,
    col_data: Tensor,
    col_indices: Tensor,
    col_indptr: Tensor,
) -> list[Tensor]:
    return [torch.empty_like(x), torch.empty_like(y), torch.empty_like(z)]


@torch.library.custom_op(
    "tensornet::message_passing_bwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(
    grad_output_x: Tensor,
    grad_output_y: Tensor,
    grad_output_z: Tensor,
    x: Tensor,
    y: Tensor,
    z: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indices: Tensor,
    row_indptr: Tensor,
    col_data: Tensor,
    col_indices: Tensor,
    col_indptr: Tensor,
) -> list[Tensor]:
    stream = get_stream(x.device)
    device = wp.device_from_torch(x.device)
    grad_x = torch.empty_like(x)
    grad_y = torch.empty_like(y)
    grad_z = torch.empty_like(z)

    grad_edge_attr = torch.zeros_like(edge_attr)

    grad_output_x_wp = wp.from_torch(grad_output_x.detach(), return_ctype=True)
    grad_output_y_wp = wp.from_torch(grad_output_y.detach(), return_ctype=True)
    grad_output_z_wp = wp.from_torch(grad_output_z.detach(), return_ctype=True)

    x_wp = wp.from_torch(x.detach(), return_ctype=True)
    y_wp = wp.from_torch(y.detach(), return_ctype=True)
    z_wp = wp.from_torch(z.detach(), return_ctype=True)

    edge_attr_wp = wp.from_torch(edge_attr.detach(), return_ctype=True)

    col_data_wp = wp.from_torch(col_data.detach(), return_ctype=True)
    col_indices_wp = wp.from_torch(col_indices.detach(), return_ctype=True)
    col_indptr_wp = wp.from_torch(col_indptr.detach(), return_ctype=True)

    grad_x_wp = wp.from_torch(grad_x.detach(), return_ctype=True)
    grad_y_wp = wp.from_torch(grad_y.detach(), return_ctype=True)
    grad_z_wp = wp.from_torch(grad_z.detach(), return_ctype=True)
    grad_edge_attr_wp = wp.from_torch(grad_edge_attr.detach(), return_ctype=True)

    message_passing_bwd = get_module("message_passing_bwd", [str(x.dtype)])

    wp.launch(
        message_passing_bwd,
        dim=(x.shape[0], x.shape[-1]),
        stream=stream,
        device=device,
        inputs=(
            x_wp,
            y_wp,
            z_wp,
            edge_attr_wp,
            grad_output_x_wp,
            grad_output_y_wp,
            grad_output_z_wp,
            col_data_wp,
            col_indices_wp,
            col_indptr_wp,
            grad_x_wp,
            grad_y_wp,
            grad_z_wp,
            grad_edge_attr_wp,
        ),
    )

    return [grad_x, grad_y, grad_z, grad_edge_attr]


@torch.library.register_fake("tensornet::message_passing_bwd_primitive")
def _(
    grad_output_x: Tensor,
    grad_output_y: Tensor,
    grad_output_z: Tensor,
    x: Tensor,
    y: Tensor,
    z: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indices: Tensor,
    row_indptr: Tensor,
    col_data: Tensor,
    col_indices: Tensor,
    col_indptr: Tensor,
) -> list[Tensor]:
    return [
        torch.empty_like(x),
        torch.empty_like(y),
        torch.empty_like(z),
        torch.empty_like(edge_attr),
    ]


@torch.library.custom_op(
    "tensornet::message_passing_bwd_bwd_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(
    grad_output_x: Tensor,
    grad_output_y: Tensor,
    grad_output_z: Tensor,
    grad_grad_x: Tensor,
    grad_grad_y: Tensor,
    grad_grad_z: Tensor,
    grad_grad_edge_attr: Tensor,
    x: Tensor,
    y: Tensor,
    z: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indices: Tensor,
    row_indptr: Tensor,
    col_data: Tensor,
    col_indices: Tensor,
    col_indptr: Tensor,
) -> list[Tensor]:
    stream = get_stream(x.device)
    device = wp.device_from_torch(x.device)

    # Convert inputs to warp arrays
    x_wp = wp.from_torch(x.detach(), return_ctype=True)
    y_wp = wp.from_torch(y.detach(), return_ctype=True)
    z_wp = wp.from_torch(z.detach(), return_ctype=True)
    edge_attr_wp = wp.from_torch(edge_attr.detach(), return_ctype=True)

    grad_grad_x_wp = wp.from_torch(grad_grad_x.detach(), return_ctype=True)
    grad_grad_y_wp = wp.from_torch(grad_grad_y.detach(), return_ctype=True)
    grad_grad_z_wp = wp.from_torch(grad_grad_z.detach(), return_ctype=True)
    grad_grad_edge_attr_wp = wp.from_torch(
        grad_grad_edge_attr.detach(), return_ctype=True
    )

    grad_output_x_wp = wp.from_torch(grad_output_x.detach(), return_ctype=True)
    grad_output_y_wp = wp.from_torch(grad_output_y.detach(), return_ctype=True)
    grad_output_z_wp = wp.from_torch(grad_output_z.detach(), return_ctype=True)

    col_data_wp = wp.from_torch(col_data.detach(), return_ctype=True)
    col_indices_wp = wp.from_torch(col_indices.detach(), return_ctype=True)
    col_indptr_wp = wp.from_torch(col_indptr.detach(), return_ctype=True)

    row_data_wp = wp.from_torch(row_data.detach(), return_ctype=True)
    row_indices_wp = wp.from_torch(row_indices.detach(), return_ctype=True)
    row_indptr_wp = wp.from_torch(row_indptr.detach(), return_ctype=True)

    # Allocate output tensors (no zero-init needed with two-kernel approach)
    dgrad_x = torch.empty_like(x)
    dgrad_y = torch.empty_like(y)
    dgrad_z = torch.empty_like(z)
    dgrad_edge_attr = torch.empty_like(edge_attr)
    dgrad_output_x = torch.empty_like(grad_output_x)
    dgrad_output_y = torch.empty_like(grad_output_y)
    dgrad_output_z = torch.empty_like(grad_output_z)

    dgrad_x_wp = wp.from_torch(dgrad_x.detach(), return_ctype=True)
    dgrad_y_wp = wp.from_torch(dgrad_y.detach(), return_ctype=True)
    dgrad_z_wp = wp.from_torch(dgrad_z.detach(), return_ctype=True)
    dgrad_edge_attr_wp = wp.from_torch(dgrad_edge_attr.detach(), return_ctype=True)
    dgrad_output_x_wp = wp.from_torch(dgrad_output_x.detach(), return_ctype=True)
    dgrad_output_y_wp = wp.from_torch(dgrad_output_y.detach(), return_ctype=True)
    dgrad_output_z_wp = wp.from_torch(dgrad_output_z.detach(), return_ctype=True)

    # Kernel 1: col-based - computes d2I, d2A, d2S, d2edge_attr
    message_passing_edge_bwd_bwd = get_module(
        "message_passing_edge_bwd_bwd", [str(x.dtype)]
    )
    wp.launch(
        message_passing_edge_bwd_bwd,
        dim=(x.shape[0], x.shape[-1]),
        stream=stream,
        device=device,
        inputs=(
            x_wp,
            y_wp,
            z_wp,
            grad_grad_x_wp,
            grad_grad_y_wp,
            grad_grad_z_wp,
            grad_grad_edge_attr_wp,
            grad_output_x_wp,
            grad_output_y_wp,
            grad_output_z_wp,
            col_data_wp,
            col_indices_wp,
            col_indptr_wp,
            dgrad_x_wp,
            dgrad_y_wp,
            dgrad_z_wp,
            dgrad_edge_attr_wp,
        ),
    )

    # Kernel 2: row-based - computes d2output_I, d2output_A, d2output_S
    message_passing_output_bwd_bwd = get_module(
        "message_passing_output_bwd_bwd", [str(x.dtype)]
    )
    wp.launch(
        message_passing_output_bwd_bwd,
        dim=(x.shape[0], x.shape[-1]),
        stream=stream,
        device=device,
        inputs=(
            x_wp,
            y_wp,
            z_wp,
            edge_attr_wp,
            grad_grad_x_wp,
            grad_grad_y_wp,
            grad_grad_z_wp,
            grad_grad_edge_attr_wp,
            row_data_wp,
            row_indices_wp,
            row_indptr_wp,
            dgrad_output_x_wp,
            dgrad_output_y_wp,
            dgrad_output_z_wp,
        ),
    )

    return [
        dgrad_output_x,
        dgrad_output_y,
        dgrad_output_z,
        dgrad_x,
        dgrad_y,
        dgrad_z,
        dgrad_edge_attr,
    ]


@torch.library.register_fake("tensornet::message_passing_bwd_bwd_primitive")
def _(
    grad_output_x: Tensor,
    grad_output_y: Tensor,
    grad_output_z: Tensor,
    grad_grad_x: Tensor,
    grad_grad_y: Tensor,
    grad_grad_z: Tensor,
    grad_grad_edge_attr: Tensor,
    x: Tensor,
    y: Tensor,
    z: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indices: Tensor,
    row_indptr: Tensor,
    col_data: Tensor,
    col_indices: Tensor,
    col_indptr: Tensor,
) -> list[Tensor]:
    return [
        torch.empty_like(grad_output_x),
        torch.empty_like(grad_output_y),
        torch.empty_like(grad_output_z),
        torch.empty_like(grad_grad_x),
        torch.empty_like(grad_grad_y),
        torch.empty_like(grad_grad_z),
        torch.empty_like(grad_grad_edge_attr),
    ]


def message_passing_setup_fwd_context(ctx, inputs, output):
    (
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    ) = inputs
    ctx.save_for_backward(
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    )


def message_passing_setup_bwd_context(ctx, inputs, output):
    (
        grad_output_x,
        grad_output_y,
        grad_output_z,
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    ) = inputs
    ctx.save_for_backward(
        grad_output_x,
        grad_output_y,
        grad_output_z,
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    )


@torch.compiler.allow_in_graph
def message_passing_fwd(*args):
    return torch.ops.tensornet.message_passing_fwd_primitive(*args)


@torch.compiler.allow_in_graph
def message_passing_bwd(ctx, grad_outputs):
    (
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    ) = ctx.saved_tensors

    result = torch.ops.tensornet.message_passing_bwd_primitive(
        grad_outputs[0],
        grad_outputs[1],
        grad_outputs[2],
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    )

    grad_x, grad_y, grad_z, grad_edge_attr = result

    return grad_x, grad_y, grad_z, grad_edge_attr, None, None, None, None, None, None


@torch.compiler.allow_in_graph
def message_passing_bwd_bwd(ctx, *grad_outputs):
    grad_grad_x, grad_grad_y, grad_grad_z, grad_grad_edge_attr = grad_outputs[0]

    (
        grad_output_x,
        grad_output_y,
        grad_output_z,
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    ) = ctx.saved_tensors

    result = torch.ops.tensornet.message_passing_bwd_bwd_primitive(
        grad_output_x,
        grad_output_y,
        grad_output_z,
        grad_grad_x,
        grad_grad_y,
        grad_grad_z,
        grad_grad_edge_attr,
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    )

    return (
        result[0],
        result[1],
        result[2],
        result[3],
        result[4],
        result[5],
        result[6],
        None,
        None,
        None,
        None,
        None,
        None,
    )


torch.library.register_autograd(
    "tensornet::message_passing_fwd_primitive",
    message_passing_bwd,
    setup_context=message_passing_setup_fwd_context,
)

torch.library.register_autograd(
    "tensornet::message_passing_bwd_primitive",
    message_passing_bwd_bwd,
    setup_context=message_passing_setup_bwd_context,
)


def fn_message_passing(
    x: Tensor,
    y: Tensor,
    z: Tensor,
    edge_attr: Tensor,
    row_data: Tensor,
    row_indices: Tensor,
    row_indptr: Tensor,
    col_data: Tensor,
    col_indices: Tensor,
    col_indptr: Tensor,
) -> list[Tensor]:
    return torch.ops.tensornet.message_passing_fwd_primitive(
        x,
        y,
        z,
        edge_attr,
        row_data,
        row_indices,
        row_indptr,
        col_data,
        col_indices,
        col_indptr,
    )



import torch
from typing import Optional, Tuple
from torch import Tensor, nn
from torchmdnet.models.utils import (
    CosineCutoff,
    OptimizedDistance,
    rbf_class_mapping,
    act_class_mapping,
    scatter,
    MLP,
)
from warnings import warn


import torchmdnet.models.tensornet as _tn
from torchmdnet.models.tensornet import (
    _decompose_tensor,
    _compose_tensor,
    _tensor_matmul_o3,
    _tensor_matmul_so3,
    tensor_norm,
    TensorEmbedding,
    tensornet_interaction_message_passing,
)

# OPT and the warp function bindings are shared with tensornet to ensure consistency.
# tensornet.py performs the try/except detection; we import its results here.
OPT = _tn.OPT
if OPT:
    from torchmdnet.extensions.warp_ops import (
        graph_transform,
        fn_message_passing as fn_tensornet_interaction_message_passing,
        fn_compose_tensor as compose_tensor,
        fn_decompose_tensor as decompose_tensor,
        fn_tensor_matmul_o3_3x3 as tensor_matmul_o3,
        fn_tensor_matmul_so3_3x3 as tensor_matmul_so3,
        fn_tensor_norm3,
    )
else:
    compose_tensor = _compose_tensor
    decompose_tensor = _decompose_tensor
    tensor_matmul_o3 = _tensor_matmul_o3
    tensor_matmul_so3 = _tensor_matmul_so3


__all__ = ["TensorNet2"]


class ChargePredict(nn.Module):
    def __init__(self, hidden_channels, activation, q_dim=16, static_shapes=False):
        super(ChargePredict, self).__init__()
        self.q_dim = q_dim
        self.q_norm = nn.LayerNorm(3 * hidden_channels)
        self.q_mlp = MLP(3 * hidden_channels, 2 * q_dim, hidden_channels, activation, 1)
        self.static_shapes = static_shapes
        self.dim_size = 0

        # for torchsript
        self.opt = OPT

    def reset_parameters(self):
        self.q_norm.reset_parameters()
        self.q_mlp.reset_parameters()

    def mol_sum(self, x: Tensor, batch: Tensor) -> Tensor:
        # torch.compile and torch.export don't support .item() calls during tracing
        # The model should be warmed up before compilation to set the correct dim_size
        if torch.compiler.is_compiling():
            pass
        elif torch.jit.is_scripting():
            # TorchScript doesn't support torch.cuda.is_current_stream_capturing()
            # For CPU, always update dim_size (no CUDA graphs on CPU)
            # For CUDA with static_shapes, only update once (first call sets dim_size for CUDA graph capture)
            # For CUDA without static_shapes, always update (dynamic batch sizes)
            if not x.is_cuda or not self.static_shapes or self.dim_size == 0:
                self.dim_size = int(batch.max().item() + 1)
        else:
            is_capturing = x.is_cuda and torch.cuda.is_current_stream_capturing()
            if not x.is_cuda or not is_capturing:
                self.dim_size = int(batch.max().item() + 1)
            if is_capturing:
                assert (
                    self.dim_size > 0
                ), "Warming up is needed before capturing the model into a CUDA graph"
                warn(
                    "CUDA graph capture will lock the batch to the current number of samples ({}). Changing this will result in a crash".format(
                        self.dim_size
                    )
                )

        if self.dim_size == 1:
            # do the direct sum
            out = x.sum(dim=0).unsqueeze(0)
        else:
            out = torch.zeros(
                self.dim_size, x.shape[-1], device=x.device, dtype=x.dtype
            )
            out = out.index_add(0, batch, x)

        return out

    def qeq(self, old_charges, f, batch, Q) -> Tensor:

        # if we are using static shapes we can skip the last dummy atom
        if self.static_shapes and not self.opt:
            f = f[:-1]
            old_charges = old_charges[:-1]
            Q = Q[:-1]  # Q is already per atom

        f_u = f**2

        epsilon = 1.0e-6

        F_u = self.mol_sum(f_u, batch)
        F_u = F_u + epsilon
        Q_u = self.mol_sum(old_charges, batch)

        # Q is already per atom
        dQ = Q.unsqueeze(-1) - Q_u[batch]

        F_u = F_u[batch]

        _f = f_u / F_u
        new_charges = old_charges + _f * dQ

        if (
            self.static_shapes and not self.opt
        ):  # add back the dummy atom so shapes match
            new_charges = torch.cat(
                (
                    new_charges,
                    torch.zeros(
                        (1, new_charges.shape[1]),
                        device=new_charges.device,
                        dtype=new_charges.dtype,
                    ),
                )
            )

        return new_charges

    def forward(self, X, batch, Q):
        I, A, S = decompose_tensor(X)
        if not self.opt:
            _x = torch.cat((I, tensor_norm(A), tensor_norm(S)), dim=-1)
        else:
            # for this MLP we directly use the value of I not the norm.
            _AS = compose_tensor(torch.zeros_like(I), A, S)
            _AS_norm = fn_tensor_norm3(_AS)[:, I.shape[-1] :]
            _x = torch.cat((I.squeeze(1), _AS_norm), dim=-1)

        _charges_f = self.q_mlp(self.q_norm(_x))
        charges, _f = _charges_f[:, : self.q_dim], _charges_f[:, self.q_dim :]
        charges = self.qeq(charges, _f, batch, Q)

        return charges


class TensorNet2(nn.Module):
    r"""TensorNet with Neutral Charge Equilibration
    
    
    Combines AIMNet2 style neutral charge equilibration procedure with TensorNet
    
    
    TensorNet: Cartesian Tensor Representations for Efficient Learning of Molecular Potentials; G. Simeon and G. de Fabritiis.
    NeurIPS 2023.

    This function optionally supports periodic boundary conditions with arbitrary triclinic boxes.
    For a given cutoff, :math:`r_c`, the box vectors :math:`\vec{a},\vec{b},\vec{c}` must satisfy certain requirements:

    .. math::

      \begin{align*}
      a_y = a_z = b_z &= 0 \\
      a_x, b_y, c_z &\geq 2 r_c \\
      a_x &\geq 2  b_x \\
      a_x &\geq 2  c_x \\
      b_y &\geq 2  c_y
      \end{align*}

    These requirements correspond to a particular rotation of the system and reduced form of the vectors, as well as the requirement that the cutoff be no larger than half the box width.

    Args:
        hidden_channels (int, optional): Hidden embedding size.
            (default: :obj:`128`)
        q_dim (int, optional): Charge channel size.
            (default: :obj:`16`)
        num_layers (int, optional): The number of interaction layers.
            (default: :obj:`2`)
        num_rbf (int, optional): The number of radial basis functions :math:`\mu`.
            (default: :obj:`32`)
        rbf_type (string, optional): The type of radial basis function to use.
            (default: :obj:`"expnorm"`)
        trainable_rbf (bool, optional): Whether to train RBF parameters with
            backpropagation. (default: :obj:`False`)
        activation (string, optional): The type of activation function to use.
            (default: :obj:`"silu"`)
        cutoff_lower (float, optional): Lower cutoff distance for interatomic interactions.
            (default: :obj:`0.0`)
        cutoff_upper (float, optional): Upper cutoff distance for interatomic interactions.
            (default: :obj:`4.5`)
        max_z (int, optional): Maximum atomic number. Used for initializing embeddings.
            (default: :obj:`128`)
        max_num_neighbors (int, optional): Maximum number of neighbors to return for a
            given node/atom when constructing the molecular graph during forward passes.
            (default: :obj:`64`)
        equivariance_invariance_group (string, optional): Group under whose action on input
            positions internal tensor features will be equivariant and scalar predictions
            will be invariant. O(3) or SO(3).
            (default :obj:`"O(3)"`)
        box_vecs (Tensor, optional):
            The vectors defining the periodic box.  This must have shape `(3, 3)`,
            where `box_vectors[0] = a`, `box_vectors[1] = b`, and `box_vectors[2] = c`.
            If this is omitted, periodic boundary conditions are not applied.
            (default: :obj:`None`)
        static_shapes (bool, optional): Whether to enforce static shapes.
            Makes the model CUDA-graph compatible.
            (default: :obj:`True`)
        output_charges (bool, optional): Whether charges should be output. The current 
            output is done by appending to the node features x that will be passed to the output 
            model. If True the output model must be compatible with the output charges.
            (default: :obj:`False`)
    """

    def __init__(
        self,
        hidden_channels=128,
        q_dim=16,
        num_layers=2,
        num_rbf=32,
        rbf_type="expnorm",
        trainable_rbf=False,
        activation="silu",
        cutoff_lower=0,
        cutoff_upper=4.5,
        max_num_neighbors=64,
        max_z=128,
        equivariance_invariance_group="O(3)",
        static_shapes=True,
        dtype=torch.float32,
        box_vecs=None,
        output_charges=False,
    ):
        super(TensorNet2, self).__init__()

        self.q_dim = q_dim
        self.output_charges = output_charges

        assert rbf_type in rbf_class_mapping, (
            f'Unknown RBF type "{rbf_type}". '
            f'Choose from {", ".join(rbf_class_mapping.keys())}.'
        )
        assert activation in act_class_mapping, (
            f'Unknown activation function "{activation}". '
            f'Choose from {", ".join(act_class_mapping.keys())}.'
        )

        assert equivariance_invariance_group in ["O(3)", "SO(3)"], (
            f'Unknown group "{equivariance_invariance_group}". '
            f"Choose O(3) or SO(3)."
        )
        self.hidden_channels = hidden_channels
        self.equivariance_invariance_group = equivariance_invariance_group
        self.num_layers = num_layers
        self.num_rbf = num_rbf
        self.rbf_type = rbf_type
        self.activation = activation
        self.cutoff_lower = cutoff_lower
        self.cutoff_upper = cutoff_upper
        act_class = act_class_mapping[activation]
        self.distance_expansion = rbf_class_mapping[rbf_type](
            cutoff_lower, cutoff_upper, num_rbf, trainable_rbf
        )
        self.tensor_embedding = TensorEmbedding(
            hidden_channels,
            num_rbf,
            act_class,
            cutoff_lower,
            cutoff_upper,
            trainable_rbf,
            max_z,
            dtype,
        )

        self.charge_predict_0 = ChargePredict(
            hidden_channels, activation, self.q_dim, static_shapes
        )

        self.layers = nn.ModuleList()
        self.charge_predicts = nn.ModuleList()
        if num_layers != 0:
            for l in range(num_layers):
                _layer = Interaction(
                    num_rbf,
                    hidden_channels,
                    self.q_dim,
                    activation,
                    cutoff_lower,
                    cutoff_upper,
                    equivariance_invariance_group,
                    dtype,
                )
                self.layers.append(_layer)
                _charge_predict = ChargePredict(
                    hidden_channels, activation, self.q_dim, static_shapes
                )
                self.charge_predicts.append(_charge_predict)

        self.linear = nn.Linear(3 * hidden_channels, hidden_channels, dtype=dtype)
        self.out_norm = nn.LayerNorm(3 * hidden_channels, dtype=dtype)
        self.act = act_class()
        # Resize to fit set to false ensures Distance returns a statically-shaped tensor of size max_num_pairs=pos.size*max_num_neigbors
        # negative max_num_pairs argument means "per particle"
        # long_edge_index set to False saves memory and spares some kernel launches by keeping neighbor indices as int32.
        self.static_shapes = static_shapes
        self.distance = OptimizedDistance(
            cutoff_lower,
            cutoff_upper,
            max_num_pairs=-max_num_neighbors,
            return_vecs=True,
            loop=True,
            resize_to_fit=not self.static_shapes,
            box=box_vecs,
            long_edge_index=True,
        )

        # for torchscript
        self.opt = OPT

        self.reset_parameters()

    def reset_parameters(self):
        self.tensor_embedding.reset_parameters()
        for layer in self.layers:
            layer.reset_parameters()
        self.linear.reset_parameters()
        self.out_norm.reset_parameters()
        self.charge_predict_0.reset_parameters()
        for layer in self.charge_predicts:
            layer.reset_parameters()

    def forward(
        self,
        z: Tensor,
        pos: Tensor,
        batch: Tensor,
        box: Optional[Tensor] = None,
        q: Optional[Tensor] = None,
        s: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Optional[Tensor], Tensor, Tensor, Tensor]:
        # Obtain graph, with distances and relative position vectors
        edge_index, edge_weight, edge_vec = self.distance(pos, batch, box)

        Q = q  # total molecule charge

        if self.opt:
            # perpare graph indices for message passing
            row_data, row_indices, row_indptr, col_data, col_indices, col_indptr = (
                graph_transform(edge_index.int(), z.shape[0])
            )
        else:
            row_data, row_indices, row_indptr, col_data, col_indices, col_indptr = (
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
                torch.empty(0, dtype=torch.long),
            )

        # This assert convinces TorchScript that edge_vec is a Tensor and not an Optional[Tensor]
        assert (
            edge_vec is not None
        ), "Distance module did not return directional information"
        # Total charge Q is a molecule-wise property. We transform it into an atom-wise property, with all atoms belonging to the same molecule being assigned the same charge q
        if Q is None:
            Q = torch.zeros_like(z, device=z.device, dtype=z.dtype)
        else:
            Q = Q[batch]
        zp = z
        if self.static_shapes:
            mask = (edge_index[0] < 0).unsqueeze(0).expand_as(edge_index)
            zp = torch.cat((z, torch.zeros(1, device=z.device, dtype=z.dtype)), dim=0)

            if not self.opt:
                Q = torch.cat(
                    (Q, torch.zeros(1, device=Q.device, dtype=Q.dtype)), dim=0
                )
            # I trick the model into thinking that the masked edges pertain to the extra atom
            # WARNING: This can hurt performance if max_num_pairs >> actual_num_pairs
            edge_index = edge_index.masked_fill(mask, z.shape[0])
            edge_weight = edge_weight.masked_fill(mask[0], 0)
            edge_vec = edge_vec.masked_fill(
                mask[0].unsqueeze(-1).expand_as(edge_vec), 0
            )
        edge_attr = self.distance_expansion(edge_weight)
        mask = edge_index[0] == edge_index[1]
        # Normalizing edge vectors by their length can result in NaNs, breaking Autograd.
        # I avoid dividing by zero by setting the weight of self edges and self loops to 1
        edge_vec = edge_vec / edge_weight.masked_fill(mask, 1).unsqueeze(1)

        graph_info = {"edge_index": edge_index}
        graph_info.update(
            {
                "row_data": row_data,
                "row_indices": row_indices,
                "row_indptr": row_indptr,
                "col_data": col_data,
                "col_indices": col_indices,
                "col_indptr": col_indptr,
            }
        )

        X = self.tensor_embedding(zp, graph_info, edge_weight, edge_vec, edge_attr)

        # from the initial embedding we compute some partial charges
        charges_0 = self.charge_predict_0(X, batch, Q)

        # we save them for the ouput layer
        charge_list = [charges_0]

        old_charges = charges_0  # intermediate so we can loop

        for l, (layer, charge_predictor) in enumerate(
            zip(self.layers, self.charge_predicts)
        ):
            # updated node tensor feratures
            X = layer(X, old_charges, graph_info, edge_weight, edge_attr)

            # predict new charges
            new_charges = charge_predictor(X, batch, Q)
            charge_list.append(new_charges)
            old_charges = new_charges

        # we will have (l+1 * qdim) charges for each node
        charges = torch.cat(charge_list, dim=-1)

        if not self.opt:
            I, A, S = decompose_tensor(X)
            x = torch.cat((3 * I**2, tensor_norm(A), tensor_norm(S)), dim=-1)
            x = self.out_norm(x)
            x = self.act(self.linear((x)))

            # append charges for coulomb in output module
            if self.output_charges:
                x = torch.cat([x, charges], dim=-1)

            # Remove the extra atom
            if self.static_shapes:
                x = x[:-1]

        else:
            x = fn_tensor_norm3(X)
            x = self.out_norm(x)
            x = self.act(self.linear((x)))

            # append charges for coulomb in output module
            if self.output_charges:
                x = torch.cat([x, charges], dim=-1)

        return x, None, z, pos, batch


class Interaction(nn.Module):
    """Interaction layer.

    :meta private:
    """

    def __init__(
        self,
        num_rbf,
        hidden_channels,
        q_dim,
        activation,
        cutoff_lower,
        cutoff_upper,
        equivariance_invariance_group,
        dtype=torch.float32,
    ):
        super(Interaction, self).__init__()

        self.hidden_channels = hidden_channels
        self.cutoff = CosineCutoff(cutoff_lower, cutoff_upper)
        self.linears_scalar = nn.ModuleList()
        self.linears_scalar.append(
            nn.Linear(num_rbf + 2 * q_dim, self.hidden_channels, bias=True, dtype=dtype)
        )
        self.linears_scalar.append(
            nn.Linear(
                self.hidden_channels, 2 * self.hidden_channels, bias=True, dtype=dtype
            )
        )
        self.linears_scalar.append(
            nn.Linear(
                2 * self.hidden_channels,
                3 * self.hidden_channels,
                bias=True,
                dtype=dtype,
            )
        )
        self.linears_tensor = nn.ModuleList()
        for _ in range(6):
            self.linears_tensor.append(
                nn.Linear(hidden_channels, self.hidden_channels, bias=False)
            )
        self.act = act_class_mapping[activation]()
        self.equivariance_invariance_group = equivariance_invariance_group
        self.opt = OPT
        self.reset_parameters()

    def reset_parameters(self):
        for linear in self.linears_scalar:
            linear.reset_parameters()
        for linear in self.linears_tensor:
            linear.reset_parameters()

    def forward(
        self,
        X: Tensor,
        charges: Tensor,
        graph_info: dict[str, Tensor],
        edge_weight: Tensor,
        edge_attr: Tensor,
    ) -> Tensor:

        C = self.cutoff(edge_weight)

        edge_index = graph_info["edge_index"]

        # add the partial charges as edge features and then do a normal tensornet update

        if self.opt:  # TODO: and static shapes
            # add a dummy atom to the charges so we can do index select correctly with static shapes
            _charges = torch.cat(
                (
                    charges,
                    torch.zeros(
                        1, charges.shape[1], device=charges.device, dtype=charges.dtype
                    ),
                ),
                dim=0,
            )

            charge_edges_i = _charges.index_select(0, edge_index[0])
            charge_edges_j = _charges.index_select(0, edge_index[1])
        else:
            charge_edges_i = charges.index_select(0, edge_index[0])
            charge_edges_j = charges.index_select(0, edge_index[1])

        x_edge_attr = torch.cat([edge_attr, charge_edges_i, charge_edges_j], dim=-1)

        for linear_scalar in self.linears_scalar:
            x_edge_attr = self.act(linear_scalar(x_edge_attr))
        x_edge_attr = (x_edge_attr * C.view(-1, 1)).reshape(
            x_edge_attr.shape[0], 3, self.hidden_channels
        )

        X = X / (tensor_norm(X) + 1)[:, None, None, :]

        I, A, S = decompose_tensor(X)

        I = self.linears_tensor[0](I)
        A = self.linears_tensor[1](A)
        S = self.linears_tensor[2](S)
        Y = compose_tensor(I, A, S)

        if not self.opt:
            edge_index = graph_info["edge_index"]

            Im, Am, Sm = tensornet_interaction_message_passing(
                I, A, S, x_edge_attr, edge_index, X.shape[0]
            )

        else:

            row_data = graph_info["row_data"]
            row_indices = graph_info["row_indices"]
            row_indptr = graph_info["row_indptr"]
            col_data = graph_info["col_data"]
            col_indices = graph_info["col_indices"]
            col_indptr = graph_info["col_indptr"]

            Im, Am, Sm = fn_tensornet_interaction_message_passing(
                I,
                A,
                S,
                x_edge_attr,
                row_data,
                row_indices,
                row_indptr,
                col_data,
                col_indices,
                col_indptr,
            )

        msg = compose_tensor(Im, Am, Sm)

        if self.equivariance_invariance_group == "O(3)":
            C = tensor_matmul_o3(Y, msg)
            I, A, S = decompose_tensor(C)
        if self.equivariance_invariance_group == "SO(3)":
            C = 2 * tensor_matmul_so3(Y, msg)
            I, A, S = decompose_tensor(C)

        normp1 = tensor_norm(C) + 1

        if not self.opt:
            I, A, S = (
                I / normp1,
                A / normp1[..., None, None, :],
                S / normp1[..., None, None, :],
            )
        else:
            I = I / normp1.unsqueeze(1)
            A = A / normp1.unsqueeze(1)
            S = S / normp1.unsqueeze(1)

        I = self.linears_tensor[3](I)
        A = self.linears_tensor[4](A)
        S = self.linears_tensor[5](S)
        dX = compose_tensor(I, A, S)
        X = X + dX + tensor_matmul_so3(dX, dX)

        return X



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import warp as wp

from .utils import add_module, get_wp_fp_dtype


def generate_message_passing(dtype: str):
    dtype_wp = get_wp_fp_dtype(dtype)

    class vec3(wp.types.vector(length=3, dtype=dtype_wp)):
        pass

    class vec5(wp.types.vector(length=5, dtype=dtype_wp)):
        pass

    def message_passing_fwd(
        I: wp.array(ndim=3, dtype=dtype_wp),
        A: wp.array(ndim=3, dtype=dtype_wp),
        S: wp.array(ndim=3, dtype=dtype_wp),
        edge_attr: wp.array(ndim=3, dtype=dtype_wp),
        row_data: wp.array(ndim=1, dtype=wp.int32),
        row_indices: wp.array(ndim=1, dtype=wp.int32),
        row_indptr: wp.array(ndim=1, dtype=wp.int32),
        output_I: wp.array(ndim=3, dtype=dtype_wp),
        output_A: wp.array(ndim=3, dtype=dtype_wp),
        output_S: wp.array(ndim=3, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        output_I_reg = I.dtype(0)
        output_A_reg = vec3(I.dtype(0))
        output_S_reg = vec5(I.dtype(0))

        for i in range(row_indptr[b], row_indptr[b + 1]):
            idx_j = row_indices[i]
            idx_w = row_data[i]
            wI = edge_attr[idx_w, 0, h]
            wA = edge_attr[idx_w, 1, h]
            wS = edge_attr[idx_w, 2, h]

            output_I_reg += I[idx_j, 0, h] * wI
            for j in range(3):
                output_A_reg[j] += A[idx_j, j, h] * wA
            for j in range(5):
                output_S_reg[j] += S[idx_j, j, h] * wS

        output_I[b, 0, h] = output_I_reg
        for j in range(3):
            output_A[b, j, h] = output_A_reg[j]

        for j in range(5):
            output_S[b, j, h] = output_S_reg[j]

    def message_passing_bwd(
        I: wp.array(ndim=3, dtype=dtype_wp),
        A: wp.array(ndim=3, dtype=dtype_wp),
        S: wp.array(ndim=3, dtype=dtype_wp),
        edge_attr: wp.array(ndim=3, dtype=dtype_wp),
        doutput_I: wp.array(ndim=3, dtype=dtype_wp),
        doutput_A: wp.array(ndim=3, dtype=dtype_wp),
        doutput_S: wp.array(ndim=3, dtype=dtype_wp),
        col_data: wp.array(ndim=1, dtype=wp.int32),
        col_indices: wp.array(ndim=1, dtype=wp.int32),
        col_indptr: wp.array(ndim=1, dtype=wp.int32),
        dI: wp.array(ndim=3, dtype=dtype_wp),
        dA: wp.array(ndim=3, dtype=dtype_wp),
        dS: wp.array(ndim=3, dtype=dtype_wp),
        dedge_attr: wp.array(ndim=3, dtype=dtype_wp),
    ):
        b, h = wp.tid()

        dI_reg = I.dtype(0.0)
        dA_reg = vec3(I.dtype(0.0))
        dS_reg = vec5(I.dtype(0.0))

        for i in range(col_indptr[b], col_indptr[b + 1]):
            idx_j = col_indices[i]
            idx_w = col_data[i]

            wI = edge_attr[idx_w, 0, h]
            doutput_I_j = doutput_I[idx_j, 0, h]
            dI_reg += doutput_I_j * wI
            dedge_attr[idx_w, 0, h] = doutput_I_j * I[b, 0, h]

            # A
            wA = edge_attr[idx_w, 1, h]
            dweight_A = I.dtype(0.0)
            for j in range(3):
                dA_reg[j] += doutput_A[idx_j, j, h] * wA
                dweight_A += doutput_A[idx_j, j, h] * A[b, j, h]
            dedge_attr[idx_w, 1, h] = dweight_A

            # S
            wS = edge_attr[idx_w, 2, h]
            dweight_S = I.dtype(0.0)
            for j in range(5):
                dS_reg[j] += doutput_S[idx_j, j, h] * wS
                dweight_S += doutput_S[idx_j, j, h] * S[b, j, h]
            dedge_attr[idx_w, 2, h] = dweight_S

        dI[b, 0, h] = dI_reg
        for j in range(3):
            dA[b, j, h] = dA_reg[j]
        for j in range(5):
            dS[b, j, h] = dS_reg[j]

    def message_passing_edge_bwd_bwd(
        I: wp.array(ndim=3, dtype=dtype_wp),
        A: wp.array(ndim=3, dtype=dtype_wp),
        S: wp.array(ndim=3, dtype=dtype_wp),
        dI: wp.array(ndim=3, dtype=dtype_wp),
        dA: wp.array(ndim=3, dtype=dtype_wp),
        dS: wp.array(ndim=3, dtype=dtype_wp),
        dedge_attr: wp.array(ndim=3, dtype=dtype_wp),
        doutput_I: wp.array(ndim=3, dtype=dtype_wp),
        doutput_A: wp.array(ndim=3, dtype=dtype_wp),
        doutput_S: wp.array(ndim=3, dtype=dtype_wp),
        col_data: wp.array(ndim=1, dtype=wp.int32),
        col_indices: wp.array(ndim=1, dtype=wp.int32),
        col_indptr: wp.array(ndim=1, dtype=wp.int32),
        d2I: wp.array(ndim=3, dtype=dtype_wp),
        d2A: wp.array(ndim=3, dtype=dtype_wp),
        d2S: wp.array(ndim=3, dtype=dtype_wp),
        d2edge_attr: wp.array(ndim=3, dtype=dtype_wp),
    ):
        # Col-based iteration: b is source node, idx_j is destination node
        # Computes d2I, d2A, d2S, d2edge_attr - no atomics needed
        b, h = wp.tid()

        d2I_reg = I.dtype(0)
        d2A_reg = vec3(I.dtype(0))
        d2S_reg = vec5(I.dtype(0))

        for i in range(col_indptr[b], col_indptr[b + 1]):
            idx_j = col_indices[i]  # Destination node
            idx_w = col_data[i]

            dweight_I = dedge_attr[idx_w, 0, h]
            dweight_A = dedge_attr[idx_w, 1, h]
            dweight_S = dedge_attr[idx_w, 2, h]

            # d2I[b] = Σ dedge_attr[edge] * doutput_I[dst]
            d2I_reg += doutput_I[idx_j, 0, h] * dweight_I

            # d2edge_attr[edge] = dI[src] * doutput_I[dst]
            d2edge_attr[idx_w, 0, h] = doutput_I[idx_j, 0, h] * dI[b, 0, h]

            # A
            dweight_A_reg = I.dtype(0.0)
            for j in range(3):
                d2A_reg[j] += doutput_A[idx_j, j, h] * dweight_A
                dweight_A_reg += doutput_A[idx_j, j, h] * dA[b, j, h]
            d2edge_attr[idx_w, 1, h] = dweight_A_reg

            # S
            dweight_S_reg = I.dtype(0.0)
            for j in range(5):
                d2S_reg[j] += doutput_S[idx_j, j, h] * dweight_S
                dweight_S_reg += doutput_S[idx_j, j, h] * dS[b, j, h]
            d2edge_attr[idx_w, 2, h] = dweight_S_reg

        d2I[b, 0, h] = d2I_reg

        for j in range(3):
            d2A[b, j, h] = d2A_reg[j]

        for j in range(5):
            d2S[b, j, h] = d2S_reg[j]

    def message_passing_output_bwd_bwd(
        I: wp.array(ndim=3, dtype=dtype_wp),
        A: wp.array(ndim=3, dtype=dtype_wp),
        S: wp.array(ndim=3, dtype=dtype_wp),
        edge_attr: wp.array(ndim=3, dtype=dtype_wp),
        dI: wp.array(ndim=3, dtype=dtype_wp),
        dA: wp.array(ndim=3, dtype=dtype_wp),
        dS: wp.array(ndim=3, dtype=dtype_wp),
        dedge_attr: wp.array(ndim=3, dtype=dtype_wp),
        row_data: wp.array(ndim=1, dtype=wp.int32),
        row_indices: wp.array(ndim=1, dtype=wp.int32),
        row_indptr: wp.array(ndim=1, dtype=wp.int32),
        d2output_I: wp.array(ndim=3, dtype=dtype_wp),
        d2output_A: wp.array(ndim=3, dtype=dtype_wp),
        d2output_S: wp.array(ndim=3, dtype=dtype_wp),
    ):
        # Row-based iteration: b is destination node, idx_j is source node
        # Computes d2output_I, d2output_A, d2output_S - no atomics needed
        b, h = wp.tid()

        d2output_I_reg = I.dtype(0)
        d2output_A_reg = vec3(I.dtype(0))
        d2output_S_reg = vec5(I.dtype(0))

        for i in range(row_indptr[b], row_indptr[b + 1]):
            idx_j = row_indices[i]  # Source node
            idx_w = row_data[i]

            wI = edge_attr[idx_w, 0, h]
            wA = edge_attr[idx_w, 1, h]
            wS = edge_attr[idx_w, 2, h]

            dweight_I = dedge_attr[idx_w, 0, h]
            dweight_A = dedge_attr[idx_w, 1, h]
            dweight_S = dedge_attr[idx_w, 2, h]

            # d2output_I[b] = Σ (dI[src] * edge_attr + I[src] * dedge_attr)
            d2output_I_reg += dI[idx_j, 0, h] * wI
            d2output_I_reg += I[idx_j, 0, h] * dweight_I

            # A
            for j in range(3):
                d2output_A_reg[j] += dA[idx_j, j, h] * wA
                d2output_A_reg[j] += A[idx_j, j, h] * dweight_A

            # S
            for j in range(5):
                d2output_S_reg[j] += dS[idx_j, j, h] * wS
                d2output_S_reg[j] += S[idx_j, j, h] * dweight_S

        d2output_I[b, 0, h] = d2output_I_reg

        for j in range(3):
            d2output_A[b, j, h] = d2output_A_reg[j]

        for j in range(5):
            d2output_S[b, j, h] = d2output_S_reg[j]

    return (
        wp.Kernel(
            message_passing_fwd,
            key=f"message_passing_fwd_{dtype}",
            module=wp.get_module(f"message_passing_fwd_{dtype}"),
        ),
        wp.Kernel(
            message_passing_bwd,
            key=f"message_passing_bwd_{dtype}",
            module=wp.get_module(f"message_passing_bwd_{dtype}"),
        ),
        wp.Kernel(
            message_passing_edge_bwd_bwd,
            key=f"message_passing_edge_bwd_bwd_{dtype}",
            module=wp.get_module(f"message_passing_edge_bwd_bwd_{dtype}"),
        ),
        wp.Kernel(
            message_passing_output_bwd_bwd,
            key=f"message_passing_output_bwd_bwd_{dtype}",
            module=wp.get_module(f"message_passing_output_bwd_bwd_{dtype}"),
        ),
    )


(
    message_passing_fwd_fp64,
    message_passing_bwd_fp64,
    message_passing_edge_bwd_bwd_fp64,
    message_passing_output_bwd_bwd_fp64,
) = generate_message_passing("float64")
(
    message_passing_fwd_fp32,
    message_passing_bwd_fp32,
    message_passing_edge_bwd_bwd_fp32,
    message_passing_output_bwd_bwd_fp32,
) = generate_message_passing("float32")
(
    message_passing_fwd_fp16,
    message_passing_bwd_fp16,
    message_passing_edge_bwd_bwd_fp16,
    message_passing_output_bwd_bwd_fp16,
) = generate_message_passing("float16")

add_module("message_passing_fwd", ["float64"], message_passing_fwd_fp64)
add_module("message_passing_bwd", ["float64"], message_passing_bwd_fp64)
add_module(
    "message_passing_edge_bwd_bwd", ["float64"], message_passing_edge_bwd_bwd_fp64
)
add_module(
    "message_passing_output_bwd_bwd", ["float64"], message_passing_output_bwd_bwd_fp64
)

add_module("message_passing_fwd", ["float32"], message_passing_fwd_fp32)
add_module("message_passing_bwd", ["float32"], message_passing_bwd_fp32)
add_module(
    "message_passing_edge_bwd_bwd", ["float32"], message_passing_edge_bwd_bwd_fp32
)
add_module(
    "message_passing_output_bwd_bwd", ["float32"], message_passing_output_bwd_bwd_fp32
)

add_module("message_passing_fwd", ["float16"], message_passing_fwd_fp16)
add_module("message_passing_bwd", ["float16"], message_passing_bwd_fp16)
add_module(
    "message_passing_edge_bwd_bwd", ["float16"], message_passing_edge_bwd_bwd_fp16
)
add_module(
    "message_passing_output_bwd_bwd", ["float16"], message_passing_output_bwd_bwd_fp16
)



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)
import triton
import triton.language as tl
import torch
from torch import Tensor
from typing import Tuple
from torchmdnet.extensions.triton_neighbors import (
    TritonNeighborAutograd,
    neighbor_grad_positions,
    neighbor_op_setup_context,
)
from torch.library import triton_op, wrap_triton


def _get_cell_dimensions(
    box_x: torch.float32,
    box_y: torch.float32,
    box_z: torch.float32,
    cutoff_upper: torch.float32,
) -> int:
    nx = torch.floor(box_x / cutoff_upper).clamp(min=3).long()
    ny = torch.floor(box_y / cutoff_upper).clamp(min=3).long()
    nz = torch.floor(box_z / cutoff_upper).clamp(min=3).long()
    return torch.stack([nx, ny, nz])


@triton.jit
def _tl_round(x):
    return tl.where(x >= 0, tl.math.floor(x + 0.5), tl.math.ceil(x - 0.5))


@triton.jit
def cell_neighbor_kernel(
    # Cell data structure (1D sorted approach)
    SortedIndices,  # [n_atoms] - original atom indices, sorted by cell
    SortedPositions,  # [n_atoms, 3] - positions sorted by cell (for coalesced access)
    SortedBatch,  # [n_atoms] - batch indices sorted by cell
    CellStart,  # [num_cells] - start index in sorted arrays for each cell
    CellEnd,  # [num_cells] - end index (exclusive) in sorted arrays for each cell
    # Box parameters
    BoxSizes,  # [3] - box dimensions
    CellDims,  # [3] - number of cells in each dimension (int32)
    # Output
    OutPairs,  # [2, max_pairs]
    OutDeltas,  # [max_pairs, 3]
    OutDists,  # [max_pairs]
    GlobalCounter,  # [1]
    # Scalar parameters
    max_pairs,
    cutoff_lower_sq,
    cutoff_upper_sq,
    # Flags
    use_periodic: tl.constexpr,
    loop: tl.constexpr,
    include_transpose: tl.constexpr,
    # Batch size for vectorized processing
    BATCH_SIZE: tl.constexpr,  # e.g., 32 -> processes 32×32=1024 pairs per iteration
):
    """
    Each program processes one cell (the "home" cell).
    Uses 1D sorted array with cell_start/cell_end pointers.

    Vectorized batched processing:
    - Loads BATCH_SIZE atoms at a time for both home and neighbor
    - Computes BATCH_SIZE × BATCH_SIZE distance matrix per iteration
    - Uses while loops to iterate only over actual atoms
    - Minimal waste: only last partial batch may have masked elements

    To avoid double-counting:
    - For half list (include_transpose=False): only emit pairs where home_atom > neighbor_atom
    - For full list (include_transpose=True): emit both directions
    """
    home_cell_id = tl.program_id(0)

    # Load box and cell dimensions
    box_x = tl.load(BoxSizes + 0)
    box_y = tl.load(BoxSizes + 1)
    box_z = tl.load(BoxSizes + 2)

    num_cells_x = tl.load(CellDims + 0)
    num_cells_y = tl.load(CellDims + 1)
    num_cells_z = tl.load(CellDims + 2)

    # Decompose home cell ID into 3D coordinates
    cells_yz = num_cells_y * num_cells_z
    home_cx = home_cell_id // cells_yz
    home_cy = (home_cell_id % cells_yz) // num_cells_z
    home_cz = home_cell_id % num_cells_z

    # Load home cell boundaries
    home_start = tl.load(CellStart + home_cell_id)
    home_end = tl.load(CellEnd + home_cell_id)

    # Loop over 27 neighbor cells
    for neighbor_offset in tl.range(0, 27):
        # Decompose neighbor_offset into di, dj, dk (each in {-1, 0, 1})
        di = (neighbor_offset % 3) - 1
        dj = ((neighbor_offset // 3) % 3) - 1
        dk = (neighbor_offset // 9) - 1

        # Compute neighbor cell coordinates
        ni = home_cx + di
        nj = home_cy + dj
        nk = home_cz + dk

        # Handle boundary conditions
        if use_periodic:
            ni = (ni + num_cells_x) % num_cells_x
            nj = (nj + num_cells_y) % num_cells_y
            nk = (nk + num_cells_z) % num_cells_z
            cell_valid = True
        else:
            cell_valid = (
                (ni >= 0)
                & (ni < num_cells_x)
                & (nj >= 0)
                & (nj < num_cells_y)
                & (nk >= 0)
                & (nk < num_cells_z)
            )

        neighbor_cell_id = ni * cells_yz + nj * num_cells_z + nk

        # Load neighbor cell boundaries
        neighbor_start = tl.load(CellStart + neighbor_cell_id)
        neighbor_end = tl.load(CellEnd + neighbor_cell_id)

        # If cell is invalid (non-periodic boundary), make it empty
        neighbor_start = tl.where(cell_valid, neighbor_start, 0)
        neighbor_end = tl.where(cell_valid, neighbor_end, 0)

        # Batched iteration over home atoms
        home_batch_start = home_start
        while home_batch_start < home_end:
            # Load BATCH_SIZE home atoms
            home_offsets = tl.arange(0, BATCH_SIZE)
            home_global_idx = home_batch_start + home_offsets
            home_mask = home_global_idx < home_end

            # Load home atom original indices (for output pair indices)
            home_atoms = tl.load(
                SortedIndices + home_global_idx, mask=home_mask, other=0
            )

            # Load home atom positions (sequential access - coalesced!)
            home_x = tl.load(
                SortedPositions + home_global_idx * 3 + 0, mask=home_mask, other=0.0
            )
            home_y = tl.load(
                SortedPositions + home_global_idx * 3 + 1, mask=home_mask, other=0.0
            )
            home_z = tl.load(
                SortedPositions + home_global_idx * 3 + 2, mask=home_mask, other=0.0
            )
            home_batch = tl.load(
                SortedBatch + home_global_idx, mask=home_mask, other=-1
            )

            # Batched iteration over neighbor atoms
            neighbor_batch_start = neighbor_start
            while neighbor_batch_start < neighbor_end:
                # Load BATCH_SIZE neighbor atoms
                neighbor_offsets = tl.arange(0, BATCH_SIZE)
                neighbor_global_idx = neighbor_batch_start + neighbor_offsets
                neighbor_mask = neighbor_global_idx < neighbor_end

                # Load neighbor atom original indices (for output pair indices)
                neighbor_atoms = tl.load(
                    SortedIndices + neighbor_global_idx, mask=neighbor_mask, other=0
                )

                # Load neighbor atom positions (sequential access - coalesced!)
                neighbor_x = tl.load(
                    SortedPositions + neighbor_global_idx * 3 + 0,
                    mask=neighbor_mask,
                    other=0.0,
                )
                neighbor_y = tl.load(
                    SortedPositions + neighbor_global_idx * 3 + 1,
                    mask=neighbor_mask,
                    other=0.0,
                )
                neighbor_z = tl.load(
                    SortedPositions + neighbor_global_idx * 3 + 2,
                    mask=neighbor_mask,
                    other=0.0,
                )
                neighbor_batch_vals = tl.load(
                    SortedBatch + neighbor_global_idx, mask=neighbor_mask, other=-2
                )

                # Compute pairwise distances: [BATCH_SIZE, BATCH_SIZE]
                dx = home_x[:, None] - neighbor_x[None, :]
                dy = home_y[:, None] - neighbor_y[None, :]
                dz = home_z[:, None] - neighbor_z[None, :]

                # Apply PBC
                if use_periodic:
                    dx = dx - box_x * _tl_round(dx / box_x)
                    dy = dy - box_y * _tl_round(dy / box_y)
                    dz = dz - box_z * _tl_round(dz / box_z)

                dist_sq = dx * dx + dy * dy + dz * dz

                # Build validity mask
                # 1. Distance within cutoff
                cond_dist = (dist_sq < cutoff_upper_sq) & (dist_sq >= cutoff_lower_sq)

                # 2. Same batch
                cond_batch = home_batch[:, None] == neighbor_batch_vals[None, :]

                # 3. Index ordering to avoid double-counting
                home_atoms_bc = home_atoms[:, None]
                neighbor_atoms_bc = neighbor_atoms[None, :]

                if include_transpose:
                    if loop:
                        cond_idx = True
                    else:
                        cond_idx = home_atoms_bc != neighbor_atoms_bc
                else:
                    if loop:
                        cond_idx = home_atoms_bc >= neighbor_atoms_bc
                    else:
                        cond_idx = home_atoms_bc > neighbor_atoms_bc

                # 4. Both atoms must be valid (within actual cell bounds)
                cond_valid = home_mask[:, None] & neighbor_mask[None, :]

                # Combined validity
                valid_mask = cond_dist & cond_batch & cond_idx & cond_valid

                # Count and store valid pairs
                num_found = tl.sum(valid_mask.to(tl.int32))

                if num_found > 0:
                    # Atomically reserve space in output
                    current_offset = tl.atomic_add(GlobalCounter, num_found)

                    if current_offset + num_found <= max_pairs:
                        # Compute storage indices using cumsum
                        flat_mask = tl.ravel(valid_mask)
                        csum = tl.cumsum(flat_mask.to(tl.int32), axis=0)
                        store_idx = current_offset + csum - 1

                        # Prepare flattened data
                        flat_home = tl.ravel(
                            tl.broadcast_to(
                                home_atoms[:, None], (BATCH_SIZE, BATCH_SIZE)
                            )
                        )
                        flat_neighbor = tl.ravel(
                            tl.broadcast_to(
                                neighbor_atoms[None, :], (BATCH_SIZE, BATCH_SIZE)
                            )
                        )
                        flat_dx = tl.ravel(dx)
                        flat_dy = tl.ravel(dy)
                        flat_dz = tl.ravel(dz)
                        flat_dist = tl.sqrt(tl.ravel(dist_sq))

                        # Store pairs
                        tl.store(
                            OutPairs + 0 * max_pairs + store_idx,
                            flat_home,
                            mask=flat_mask,
                        )
                        tl.store(
                            OutPairs + 1 * max_pairs + store_idx,
                            flat_neighbor,
                            mask=flat_mask,
                        )

                        # Store deltas
                        tl.store(OutDeltas + store_idx * 3 + 0, flat_dx, mask=flat_mask)
                        tl.store(OutDeltas + store_idx * 3 + 1, flat_dy, mask=flat_mask)
                        tl.store(OutDeltas + store_idx * 3 + 2, flat_dz, mask=flat_mask)

                        # Store distances
                        tl.store(OutDists + store_idx, flat_dist, mask=flat_mask)

                neighbor_batch_start += BATCH_SIZE
            home_batch_start += BATCH_SIZE


def build_cell_list(
    positions: Tensor,
    batch: Tensor,
    box_sizes: Tensor,  # [3] diagonal elements
    use_periodic: bool,
    cell_dims: Tensor,  # [3] number of cells in each dimension
    num_cells: int,  # total number of cells (fixed for CUDA graphs)
) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
    """
    Build the cell list data structure using 1D sorted arrays.

    Args:
        positions: [N, 3] atom positions
        batch: [N] batch indices
        box_sizes: [3] box diagonal elements
        use_periodic: whether to use periodic boundary conditions
        cell_dims: [3] number of cells in each dimension (pre-computed)
        num_cells: total number of cells (pre-computed, fixed for CUDA graphs)

    Returns:
        sorted_indices: [n_atoms] - original atom indices, sorted by cell
        sorted_positions: [n_atoms, 3] - positions sorted by cell (for coalesced access)
        sorted_batch: [n_atoms] - batch indices sorted by cell
        cell_start: [num_cells] - start index in sorted arrays for each cell
        cell_end: [num_cells] - end index (exclusive) in sorted arrays for each cell
    """
    device = positions.device
    n_atoms = positions.size(0)

    # Compute cell index for each atom
    if use_periodic:
        # Wrap to [0, box)
        inv_box = 1.0 / box_sizes
        wrapped = positions - torch.floor(positions * inv_box) * box_sizes
    else:
        # Shift by half box (like CUDA implementation)
        wrapped = positions + 0.5 * box_sizes

    # Cell coordinates
    cell_size = box_sizes / cell_dims.float()
    cell_coords = (wrapped / cell_size).long()
    cell_coords = torch.clamp(
        cell_coords, min=torch.zeros(3, device=device), max=cell_dims - 1
    )

    # Flat cell index
    cell_idx = (
        cell_coords[:, 0] * (cell_dims[1] * cell_dims[2])
        + cell_coords[:, 1] * cell_dims[2]
        + cell_coords[:, 2]
    ).long()

    # Sort atoms by cell index
    sorted_cell_idx, sort_order = torch.sort(cell_idx)
    sorted_indices = sort_order.int()  # Original atom indices, now sorted by cell

    # Create sorted positions and batch for coalesced memory access
    sorted_positions = positions.index_select(0, sort_order).contiguous()
    sorted_batch = batch.index_select(0, sort_order).contiguous()

    # Count atoms per cell
    cell_counts = torch.zeros(num_cells, dtype=torch.int32, device=device)
    cell_counts.scatter_add_(
        0, cell_idx, torch.ones(n_atoms, dtype=torch.int32, device=device)
    )

    # Compute cell_start and cell_end using cumsum
    cell_end = torch.cumsum(cell_counts, dim=0).int()
    cell_start = torch.zeros(num_cells, dtype=torch.int32, device=device)
    cell_start[1:] = cell_end[:-1]

    return sorted_indices, sorted_positions, sorted_batch, cell_start, cell_end


@triton_op("torchmdnet::triton_neighbor_cell", mutates_args={})
def triton_neighbor_cell(
    positions: Tensor,
    batch: Tensor,
    box_vectors: Tensor,
    use_periodic: bool,
    cutoff_lower: float,
    cutoff_upper: float,
    max_num_pairs: int,
    loop: bool,
    include_transpose: bool,
    num_cells: int,
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    device = positions.device
    dtype = positions.dtype
    n_atoms = positions.size(0)

    # Validate inputs
    if positions.dim() != 2 or positions.size(1) != 3:
        raise ValueError('Expected "positions" to have shape (N, 3)')
    if batch.dim() != 1 or batch.size(0) != n_atoms:
        raise ValueError('Expected "batch" to have shape (N,)')

    # Extract box diagonal
    box_vectors = box_vectors.contiguous()
    if box_vectors.dim() == 3:
        box_diag = box_vectors[0]
    else:
        box_diag = box_vectors
    box_sizes = torch.stack(
        [box_diag[0, 0], box_diag[1, 1], box_diag[2, 2]]
    ).contiguous()

    # Compute cell dimensions using shared utility (stays on GPU)
    cell_dims = _get_cell_dimensions(
        box_sizes[0], box_sizes[1], box_sizes[2], cutoff_upper
    )

    # Build cell list (1D sorted approach with sorted positions for coalesced access)
    sorted_indices, sorted_positions, sorted_batch, cell_start, cell_end = (
        build_cell_list(positions, batch, box_sizes, use_periodic, cell_dims, num_cells)
    )

    # Allocate outputs
    neighbors = torch.full((2, max_num_pairs), -1, device=device, dtype=torch.long)
    deltas = torch.zeros((max_num_pairs, 3), device=device, dtype=dtype)
    distances = torch.zeros((max_num_pairs,), device=device, dtype=dtype)
    num_pairs = torch.zeros((1,), device=device, dtype=torch.int32)

    # Launch kernel: one program per cell
    # BATCH_SIZE: process atoms in batches for vectorized compute
    # 32 is a good balance: 32×32=1024 elements fits in registers, minimal waste on partial batches
    BATCH_SIZE = 32

    grid = (num_cells,)
    wrap_triton(cell_neighbor_kernel)[grid](
        sorted_indices,
        sorted_positions,
        sorted_batch,
        cell_start,
        cell_end,
        box_sizes,
        cell_dims,
        neighbors,
        deltas,
        distances,
        num_pairs,
        max_num_pairs,
        cutoff_lower**2,
        cutoff_upper**2,
        use_periodic=use_periodic,
        loop=loop,
        include_transpose=include_transpose,
        BATCH_SIZE=BATCH_SIZE,
    )
    return neighbors, deltas, distances, num_pairs


def _cell_backward(ctx, grad_neighbors, grad_deltas, grad_distances, grad_num_pairs):
    grad_positions = neighbor_grad_positions(ctx, grad_deltas, grad_distances)
    # Same as brute backward + extra None for num_cells input
    return grad_positions, None, None, None, None, None, None, None, None, None


torch.library.register_autograd(
    "torchmdnet::triton_neighbor_cell",
    _cell_backward,
    setup_context=neighbor_op_setup_context,
)


class TritonCellNeighborAutograd(TritonNeighborAutograd):
    @staticmethod
    def forward(
        ctx,
        positions: Tensor,
        batch: Tensor,
        box_vectors: Tensor,
        use_periodic: bool,
        cutoff_lower: float,
        cutoff_upper: float,
        max_num_pairs: int,
        loop: bool,
        include_transpose: bool,
        num_cells: int,
    ):
        neighbors, deltas, distances, num_pairs = triton_neighbor_cell(
            positions,
            batch,
            box_vectors,
            use_periodic,
            cutoff_lower,
            cutoff_upper,
            max_num_pairs,
            loop,
            include_transpose,
            num_cells,
        )

        ctx.save_for_backward(neighbors, deltas, distances)
        ctx.num_atoms = positions.size(0)
        return neighbors, deltas, distances, num_pairs

    @staticmethod
    def backward(ctx, grad_neighbors, grad_deltas, grad_distances, grad_num_pairs):  # type: ignore[override]
        # Call parent backward (returns 9 values) and add None for num_cells
        parent_grads = TritonNeighborAutograd.backward(
            ctx, grad_neighbors, grad_deltas, grad_distances, grad_num_pairs
        )
        return (*parent_grads, None)
